SystemOrganization addCategory: #'Xpf-Multipoint-Core'!

!HandMorph methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 5/5/2009 22:59'!
processEvents
	"Process user input events from the local input devices."

	| evt evtBuf type hadAny handleCurrentEvent |
	ActiveEvent ifNotNil: [
		"Meaning that we were invoked from within an event response.
		Make sure z-order is up to date"
		self mouseOverHandler processMouseOver: lastMouseEvent].
	
	hadAny := false.
	[(evtBuf := Sensor nextEvent) isNil] whileFalse: [
		evt := nil.	"for unknown event types"
		handleCurrentEvent := true.
		type := evtBuf first.
		
		type = EventTypeMouse
			ifTrue: [
				evt := MultiPointerManager processSystemMouseEvent: evtBuf on: self.
				evt ifNil: [handleCurrentEvent := false]].
		type = EventTypeKeyboard 
			ifTrue: [evt := self generateKeyboardEvent: evtBuf].
		type = EventTypeDragDropFiles 
			ifTrue: [evt := self generateDropFilesEvent: evtBuf].
		"All other events are ignored"
		(type ~= EventTypeDragDropFiles and: [evt isNil]) ifTrue: [^self].
		(evt isNil not and: [handleCurrentEvent]) ifTrue: [
			MultiPointerManager pauseSystemMouse
				ifTrue: [MultiPointerManager keyboardHand enqueueEvent: evt]
				ifFalse: [
					"Finally, handle it"
					self handleEvent: evt.
					hadAny := true.

					"For better user feedback, return immediately after a mouse event has been processed."
					(evt isMouse and: [evt isMouseWheel not]) ifTrue: [^self]
				]
		]
	].
	
	"note: if we come here we didn't have any mouse events"
	mouseClickState notNil ifTrue: [
		"No mouse events during this cycle. Make sure click states time out accordingly"
		mouseClickState handleEvent: lastMouseEvent asMouseMove from: self].
	
	hadAny ifFalse: [
		"No pending events. Make sure z-order is up to date"
		self mouseOverHandler processMouseOver: lastMouseEvent]! !

HandMorph subclass: #MultiPointerHandMorph
	instanceVariableNames: 'queue lastPosition buttons mouseID'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xpf-Multipoint-Core'!

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 4/29/2009 17:49'!
buttons

	^ buttons! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 4/29/2009 17:50'!
buttons: newButtons

	buttons := newButtons! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 4/29/2009 17:52'!
createMorphicButtonEvents: aState

	| result |
	result := OrderedCollection new.
	0 to: 2 do: [:each |
		(aState buttons anyMask: (1 bitShift: each*2)) ifTrue: [
			self buttons: (self buttons bitOr: (1 bitShift: (2-each))).
			result add: (MouseButtonEvent new 
				setType: #mouseDown
				position: self lastPosition
				which: (1 bitShift: (2-each))
				buttons: self buttons
				hand: self
				stamp: Time millisecondClockValue)].
		(aState buttons anyMask: (1 bitShift: (each*2)+1)) ifTrue: [
			self buttons: (self buttons bitAnd: ((1 bitShift: (2-each)) bitInvert)).
			result add: (MouseButtonEvent new 
				setType: #mouseUp
				position: self lastPosition
				which: (1 bitShift: (2-each))
				buttons: self buttons
				hand: self
				stamp: Time millisecondClockValue)].
	].
	^ result! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 5/4/2009 10:37'!
createMorphicMoveEvent: aState

	| newPosition evt tempPoint |
	tempPoint := Point x: aState x y: aState y.
	newPosition := self lastPosition + tempPoint.
	
	newPosition = self lastPosition ifTrue: [^ nil].
	
	newPosition := newPosition adhereTo: Display fullBoundingBox.
	evt := MouseMoveEvent new
		setType: #mouseMove
		startPoint: self lastPosition
		endPoint: newPosition
		trail: nil
		buttons: self buttons
		hand: self
		stamp: Time millisecondClockValue.
	self lastPosition: newPosition.
	^ evt! !

!MultiPointerHandMorph methodsFor: 'event handling' stamp: 'SR 5/5/2009 23:03'!
enqueueEvent: aState

	aState isMorphicEvent
		ifTrue: [self queue nextPut: aState]
		ifFalse: [
			self queue nextPut: (self createMorphicMoveEvent: aState).
			aState buttons isZero ifFalse: [
				self queue nextPutAll: (self createMorphicButtonEvents: aState)]].! !

!MultiPointerHandMorph methodsFor: 'initialization' stamp: 'SR 4/29/2009 17:53'!
initialize

	super initialize.
	self queue: SharedQueue new.
	self lastPosition: 0 asPoint.
	self buttons: 0.
! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 4/24/2009 13:38'!
lastPosition

	^ lastPosition! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 4/24/2009 13:38'!
lastPosition: anObject

	lastPosition := anObject! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 4/24/2009 14:32'!
mouseID

	^ mouseID! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 4/24/2009 14:32'!
mouseID: anObject

	mouseID := anObject! !

!MultiPointerHandMorph methodsFor: 'drawing' stamp: 'SR 4/24/2009 13:39'!
needsToBeDrawn

	^ true! !

!MultiPointerHandMorph methodsFor: 'event handling' stamp: 'SR 5/4/2009 16:30'!
processEvents
	"Play back the next event"
	
	| hadAny |
	
	hadAny := false.
	ActiveEvent ifNotNil: [
		"Meaning that we were invoked from within an event response.
		Make sure z-order is up to date"
		self mouseOverHandler processMouseOver: lastMouseEvent].

	[self queue isEmpty] whileFalse: [ | evt |
		evt := self queue next.
		self handleEvent: evt.
		hadAny := true].
	
	"note: if we come here we didn't have any mouse events"
	mouseClickState notNil ifTrue: [
		"No mouse events during this cycle. Make sure click states time out accordingly"
		mouseClickState handleEvent: lastMouseEvent asMouseMove from: self].
	
	hadAny ifFalse: [
		"No pending events. Make sure z-order is up to date"
		self mouseOverHandler processMouseOver: lastMouseEvent]! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 4/24/2009 13:38'!
queue

	^ queue! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 4/24/2009 13:38'!
queue: anObject

	queue := anObject! !

Object subclass: #MultiPointer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xpf-Multipoint-Core'!

MultiPointer class
	instanceVariableNames: 'eventSemaphore deviceSemaphore'!

MultiPointer class
	instanceVariableNames: 'eventSemaphore deviceSemaphore'!

!MultiPointer class methodsFor: 'accessing' stamp: 'SR 4/29/2009 18:58'!
deviceSemaphore

	deviceSemaphore ifNil: [self deviceSemaphore: Semaphore new].
	^ deviceSemaphore! !

!MultiPointer class methodsFor: 'accessing' stamp: 'SR 4/29/2009 18:58'!
deviceSemaphore: anObject

	deviceSemaphore := anObject! !

!MultiPointer class methodsFor: 'accessing' stamp: 'SR 4/29/2009 19:08'!
eventSemaphore

	eventSemaphore ifNil: [self eventSemaphore: Semaphore new].
	^ eventSemaphore! !

!MultiPointer class methodsFor: 'accessing' stamp: 'SR 4/29/2009 18:56'!
eventSemaphore: anObject

	eventSemaphore := anObject! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 5/2/2009 14:54'!
getPointerCount

	<primitive: 'primitiveMultiPointerCount' module: 'MultiPointerPlugin'>
	^ 0! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 4/29/2009 16:42'!
getPointerCurrentState: aPointerID

	| result |
	result := MultiPointerState new.
	self pvtGetPointerCurrentState: aPointerID with: result.
	^ result! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 4/29/2009 13:53'!
getPointerLively: aPointerID

	| result |
	result := self pvtGetPointerLively: aPointerID.
	^ result = 0
		ifTrue: [false]
		ifFalse: [true].! !

!MultiPointer class methodsFor: 'primitives' stamp: 'mig 4/22/2009 17:34'!
getPointerName: aPointerID

	| pointerName |
	pointerName := self pvtGetPointerName: aPointerID.
	^ pointerName truncateTo: (pointerName 
			indexOf: (Character value: 0) 
			startingAt: 0 
			ifAbsent: [pointerName size + 1]) - 1! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 5/2/2009 14:54'!
pvtGetPointerCurrentState: aPointerID with: aMultiPointerState

	<primitive: 'primitiveMultiPointerCurrentState' module: 'MultiPointerPlugin'>
	self primitiveFailed! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 5/2/2009 14:53'!
pvtGetPointerLively: aPointerID

	<primitive: 'primitiveMultiPointerLively' module: 'MultiPointerPlugin'>
	self primitiveFailed! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 5/2/2009 14:54'!
pvtGetPointerName: aPointerID

	<primitive: 'primitiveMultiPointerName' module: 'MultiPointerPlugin'>
	self primitiveFailed! !

!MultiPointer class methodsFor: 'system startup' stamp: 'SR 4/29/2009 19:00'!
registerSemaphore

	| eventSemaIndex deviceSemaIndex |
	eventSemaIndex := Smalltalk registerExternalObject: self eventSemaphore.
	deviceSemaIndex := Smalltalk registerExternalObject: self deviceSemaphore.
	
	self setPointerEventSemaphoreHandle: eventSemaIndex.
	self setPointerDeviceSemaphoreHandle: deviceSemaIndex.! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 5/2/2009 14:53'!
setPointerDeviceSemaphoreHandle: aSemaIndex

	<primitive: 'primitiveMultiPointerDeviceSemaphoreHandle' module: 'MultiPointerPlugin'>
	self primitiveFailed! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 5/2/2009 14:53'!
setPointerEventSemaphoreHandle: aSemaIndex

	<primitive: 'primitiveMultiPointerEventSemaphoreHandle' module: 'MultiPointerPlugin'>
	self primitiveFailed! !

!MultiPointer class methodsFor: 'system startup' stamp: 'SR 4/29/2009 19:02'!
unregisterSemaphore

	self setPointerEventSemaphoreHandle: 0.
	self setPointerDeviceSemaphoreHandle: 0.
	
	Smalltalk unregisterExternalObject: self eventSemaphore.
	Smalltalk unregisterExternalObject: self deviceSemaphore.
	
	self eventSemaphore: nil.
	self deviceSemaphore: nil.
! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 5/2/2009 14:53'!
useHardwareMouse

	<primitive: 'primitiveMultiPointerUseHardwareMouse' module: 'MultiPointerPlugin'>
	self primitiveFailed! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 5/2/2009 14:53'!
useSoftwareMouse

	<primitive: 'primitiveMultiPointerUseSoftwareMouse' module: 'MultiPointerPlugin'>
	self primitiveFailed! !

Object subclass: #MultiPointerManager
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xpf-Multipoint-Core'!

MultiPointerManager class
	instanceVariableNames: 'handMorphs movementProcess deviceUpdateProcess pauseSystemMouse keyboardHand'!

MultiPointerManager class
	instanceVariableNames: 'handMorphs movementProcess deviceUpdateProcess pauseSystemMouse keyboardHand'!

!MultiPointerManager class methodsFor: 'private' stamp: 'SR 5/2/2009 15:37'!
createHandMorphForMouse: aMouseID

	^ self handMorphs
			detect: [:each | each mouseID = aMouseID]
			ifNone: [ | m |
				m := MultiPointerHandMorph new mouseID: aMouseID.
				self handMorphs add: m.
				World addHand: m.
				MultiPointer getPointerCurrentState: m mouseID].! !

!MultiPointerManager class methodsFor: 'private' stamp: 'SR 4/29/2009 19:50'!
destroyHandMorphForMouse: aMouseID

	| m |
	m := self handMorphs detect: [:each | each mouseID = aMouseID] ifNone: [nil].
	m ifNotNil: [
		self handMorphs remove: m.
		World removeHand: m].
			! !

!MultiPointerManager class methodsFor: 'accessing' stamp: 'SR 4/29/2009 18:37'!
deviceUpdateProcess

	^ deviceUpdateProcess! !

!MultiPointerManager class methodsFor: 'accessing' stamp: 'SR 4/29/2009 18:37'!
deviceUpdateProcess: anObject

	deviceUpdateProcess := anObject! !

!MultiPointerManager class methodsFor: 'accessing' stamp: 'SR 4/24/2009 14:32'!
handMorphs

	^ handMorphs! !

!MultiPointerManager class methodsFor: 'accessing' stamp: 'SR 4/24/2009 14:32'!
handMorphs: anObject

	handMorphs := anObject! !

!MultiPointerManager class methodsFor: 'class initialization' stamp: 'SR 5/4/2009 10:52'!
initialize

	super initialize.
	self handMorphs: OrderedCollection new.
	self pauseSystemMouse: false.
	
	Smalltalk addToStartUpList: self.
	Smalltalk addToShutDownList: self.! !

!MultiPointerManager class methodsFor: 'accessing' stamp: 'SR 5/5/2009 22:56'!
keyboardHand

	^ keyboardHand! !

!MultiPointerManager class methodsFor: 'accessing' stamp: 'SR 5/5/2009 22:56'!
keyboardHand: anObject

	keyboardHand := anObject! !

!MultiPointerManager class methodsFor: 'accessing' stamp: 'SR 4/24/2009 14:51'!
movementProcess

	^ movementProcess! !

!MultiPointerManager class methodsFor: 'accessing' stamp: 'SR 4/24/2009 14:51'!
movementProcess: anObject

	movementProcess := anObject! !

!MultiPointerManager class methodsFor: 'accessing' stamp: 'SR 4/29/2009 19:54'!
pauseSystemMouse

	^ pauseSystemMouse! !

!MultiPointerManager class methodsFor: 'accessing' stamp: 'SR 4/29/2009 19:54'!
pauseSystemMouse: anObject

	pauseSystemMouse := anObject! !

!MultiPointerManager class methodsFor: 'as yet unclassified' stamp: 'SR 5/5/2009 22:56'!
processSystemMouseEvent: evtBuf on: aHandMorph
	
	| |
	self pauseSystemMouse ifFalse: [^ aHandMorph generateMouseEvent: evtBuf].
	
	^ nil! !

!MultiPointerManager class methodsFor: 'system startup' stamp: 'SR 4/24/2009 14:30'!
shutDown: aboutToQuit 

	MultiPointer unregisterSemaphore.
	super shutDown: aboutToQuit.! !

!MultiPointerManager class methodsFor: 'process control' stamp: 'SR 5/4/2009 10:43'!
startDeviceDiscovery

	self deviceUpdateProcess ifNotNil: [^self].
	self deviceUpdateProcess: [[
			0 to: MultiPointer getPointerCount - 1 do: [ :each |
				(MultiPointer getPointerLively: each)
					ifTrue: [ self createHandMorphForMouse: each ]
					ifFalse: [ self destroyHandMorphForMouse: each ]].
			MultiPointer deviceSemaphore wait.
		] repeat ] fork.! !

!MultiPointerManager class methodsFor: 'process control' stamp: 'SR 5/5/2009 22:57'!
startMoving

	self movementProcess ifNotNil: [^self].
	
	self keyboardHand: self handMorphs first.
	self pauseSystemMouse: true.
	
	
	MultiPointer useSoftwareMouse.
	self movementProcess: [
		[
			MultiPointer eventSemaphore wait.
			self handMorphs do: [:each |
				each enqueueEvent: (MultiPointer getPointerCurrentState: each mouseID)]
		] repeat
	] fork.! !

!MultiPointerManager class methodsFor: 'system startup' stamp: 'SR 4/24/2009 14:30'!
startUp: aboutToQuit 

	MultiPointer registerSemaphore.
	super startUp: aboutToQuit.! !

!MultiPointerManager class methodsFor: 'process control' stamp: 'SR 5/4/2009 10:42'!
stopDeviceDiscovery

	self deviceUpdateProcess ifNotNil: [
		self deviceUpdateProcess terminate.
		self deviceUpdateProcess: nil.
		1 to: MultiPointer getPointerCount do: [:each |
			self destroyHandMorphForMouse: (each-1)]].
! !

!MultiPointerManager class methodsFor: 'process control' stamp: 'SR 5/5/2009 22:57'!
stopMoving

	self movementProcess ifNotNil: [
		self movementProcess terminate.
		self movementProcess: nil].
	MultiPointer useHardwareMouse.

	Sensor eventQueue flush.
	self pauseSystemMouse: false.
	self keyboardHand: nil.
! !

Object subclass: #MultiPointerState
	instanceVariableNames: 'x y buttons'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xpf-Multipoint-Core'!

!MultiPointerState methodsFor: 'accessing' stamp: 'SR 4/24/2009 15:37'!
buttons

	^ buttons! !

!MultiPointerState methodsFor: 'accessing' stamp: 'SR 4/24/2009 15:37'!
buttons: anObject

	buttons := anObject! !

!MultiPointerState methodsFor: 'printing' stamp: 'SR 4/29/2009 16:50'!
printOn: aStream 
	"The receiver prints on aStream in terms of infix notation."

	x printOn: aStream.
	aStream nextPut: $@.
	y printOn: aStream.
	aStream nextPut: $@.
	buttons printOn: aStream.! !

!MultiPointerState methodsFor: 'accessing' stamp: 'SR 4/24/2009 15:37'!
x

	^ x! !

!MultiPointerState methodsFor: 'accessing' stamp: 'SR 4/24/2009 15:37'!
x: anObject

	x := anObject! !

!MultiPointerState methodsFor: 'accessing' stamp: 'SR 4/24/2009 15:37'!
y

	^ y! !

!MultiPointerState methodsFor: 'accessing' stamp: 'SR 4/24/2009 15:37'!
y: anObject

	y := anObject! !

Object subclass: #MultiPointerWidget
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xpf-Multipoint-Core'!

!MultiPointerWidget class methodsFor: 'as yet unclassified' stamp: 'SR 5/5/2009 23:18'!
createButton: actionSel label: labelSel
	
	| btn |
	btn := UITheme current
		newLabelIn: World
		label: labelSel. 
	btn 
		vResizing: #spaceFill; 
		hResizing: #spaceFill.
	btn on: #mouseUp send: actionSel to: self.
	^ btn
	! !

!MultiPointerWidget class methodsFor: 'as yet unclassified' stamp: 'SR 5/5/2009 23:18'!
grabKeyboardWithEvent: evt from: source

	MultiPointerManager pauseSystemMouse
		ifTrue: [MultiPointerManager keyboardHand: evt hand]! !

!MultiPointerWidget class methodsFor: 'as yet unclassified' stamp: 'SR 5/5/2009 23:14'!
open
	
	| align|
	align := AlignmentMorph newBounds: (50@50 corner: 700@80).
	align color: Color yellow.
	align addMorph: (self createButton: #grabKeyboardWithEvent:from: label: 'Grab Keyboard').
	align openInWorld.! !

!SharedQueue methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 5/4/2009 10:34'!
nextPut: value 
	"Send value through the receiver. If a Process has been suspended 
	waiting to receive a value through the receiver, allow it to proceed."

	value ifNil: [^ self].
	accessProtect
		critical: [writePosition > contentsArray size
						ifTrue: [self makeRoomAtEnd].
				 contentsArray at: writePosition put: value.
				 writePosition := writePosition + 1].
	readSynch signal.
	^value! !

!SharedQueue methodsFor: '*xpf-multipoint-core' stamp: 'SR 4/29/2009 17:24'!
nextPutAll: aCollection

	aCollection do: [:each | self nextPut: each].! !
MultiPointerManager initialize!
